from datetime import date, timedelta
from finance_sdk.models.entry import Entry


from enums import Frequency

def generate_entries(validated_data, initial_data):
    number_of_payments = validated_data.get("number_of_payments", 1)
    payment_value = validated_data.get("value") / number_of_payments
    frequency = validated_data.get("frequency")
    validated_data["frequency"] = Frequency.get_frequency(frequency=frequency)
    primary_payment = Entry.objects.create(**validated_data)
    payment_method = primary_payment.payment_method
    tax = payment_method.tax / 100 if payment_method else 0
    liquid_value = (
            payment_value - payment_value * tax
            if validated_data.get("type") == Entry.entry_type.ENTRY
            else None
        )
    primary_payment.value = payment_value
    primary_payment.liquid_value = liquid_value
    primary_payment.client_app = initial_data.get('client_app')
    primary_payment.reference_record_id = primary_payment.id
    primary_payment.save()

    if number_of_payments > 1:
            entry_payments = []
            for payment in range(2, number_of_payments + 1):
                entry_payments.append(
                    Entry(
                        description=validated_data.get("description"),
                        value=payment_value,
                        liquid_value=liquid_value,
                        category=validated_data.get("category"),
                        account=validated_data.get("account"),
                        cost_center=validated_data.get("cost_center"),
                        due_date=validated_data.get("due_date")
                        + timedelta(days=frequency * (payment - 1)),
                        payment_method=validated_data.get("payment_method"),
                        person=validated_data.get("person"),
                        entry_type=validated_data.get("entry_type"),
                        registry_classification=validated_data.get(
                            "registry_classification"
                        ),
                        number_of_payments=validated_data.get("number_of_payments"),
                        current_payment=payment,
                        frequency=primary_payment.frequency,
                        reference_record_id=primary_payment.pk,
                        client_app=initial_data.get('client_app'),
                    )
                )

            Entry.objects.bulk_create(entry_payments)
    else:
        received_value = validated_data.get("received_value", 0)
        total_value = validated_data.get("value", 0)

            
    if received_value < total_value and received_value > 0:
                Entry.objects.create(
                    description=validated_data.get("description"),
                    value=total_value - received_value,
                    category=validated_data.get("category"),
                    account=validated_data.get("account"),
                    cost_center=validated_data.get("cost_center"),
                    due_date=validated_data.get("due_date"),
                    person=validated_data.get("person"),
                    entry_type=validated_data.get("entry_type"),
                    registry_classification=validated_data.get("registry_classification"),
                    number_of_payments=validated_data.get("numero_parcelas", 1),
                    cliente_app_id=initial_data.get('client_app'),
                )

    return primary_payment
    






class Type:
        ENTRY = "E"
        OUT = "O"
        AUTOMATIC_PAYMENT = "AP"
        PAID_ENTRY = 'PE'

        CHOICES = (
               (ENTRY , "Entry"),
               (OUT , "Out"),
               (AUTOMATIC_PAYMENT, "Automatic Payment"),
               (PAID_ENTRY , "Paid entry")
        )


        
class RegistryClassification:
        VARIABLE = "variable"
        FIX = "fix"
        INVESTIMENT = "investiment"

        CHOICES = (
               (VARIABLE, "variable"),
               (FIX , "fix"),
               (INVESTIMENT, "investiment"),
        )


class Frequency:
        NONE = 0  
        WEEK = 7  
        BIWEEK = 15  
        MONTH = 30  
        YEAR = 365  

        CHOICES = (
               (NONE, "None"),
               (WEEK , "Week"),
               (BIWEEK, "Biweek"),
               (BIWEEK, "Biweek"),
        )

        @staticmethod
        def get_frequency(frequency: int):
            mapping = {
                Frequency.WEEK : 2,
                Frequency.BIWEEK : 3,
                Frequency.MONTH : 4,
                Frequency.YEAR : 5,
                 
            }
            return mapping.get(frequency, 1)



from django.forms import ValidationError



def raise_error(message):
    raise ValidationError(
        {
            "message": message,
        })


def validate_item_in_list(
         item, list, message, message_if_empty="", required=False
    ):
        if required and not item:
            raise_error(message_if_empty)

        if item:
            if isinstance(item, list):
                itens = item

                return (
                    itens
                    if all(item in list for item in itens)
                    else raise_error(message)
                )

            return item if item in list else raise_error(message)

        return False



def validate_natural_number(
         number, message, message_if_empty="", required=False
    ):
        if isinstance(number, list):
            number = list(filter(lambda x: x.strip() != "", number))

        elif isinstance(number, str) and number.strip() == "":
            number = None

        if required and not number:
            raise_error(message_if_empty)

        if number:
            if isinstance(number, list):
                numbers = number
                to_int = lambda x: int(x)

                return (
                    list(map(to_int, numbers))
                    if all(number.isnumeric() and number != "0" for number in numbers)
                    else  raise_error(message)
                )

            return (
                number
                if number.isnumeric() and number != "0"
                else raise_error(message)
            )

        return False



from rest_framework import serializers
from models.bank import Bank


class BankSerializer(serializers.ModelSerializer):

    class Meta:
        model = Bank
        fields = ["id", "name"]

    def validate_name(self, value):
        if not (self.instance and value == self.instance.name):
            if Bank.objects.filter(
                name=value, deactivate =False, client_app=self.initial_data.get('client_app')
            ).exists():
                raise serializers.ValidationError(
                    "Already exists a bank with this name"
                )

        return value



from django.db import models

from .category import Category
from .payment import PaymentMethod
from .accounts import Account

from ..enums import Frequency, Type, RegistryClassification

class Entry(models.Model):
    due_date = models.DateField(blank=False)
    payment_date = models.DateField(blank=True, null=True)
    description = models.TextField("Description", blank=False)
    category = models.ForeignKey(
        Category,
        verbose_name="Category",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
    )
    value = models.DecimalField(max_digits=9, decimal_places=2, blank=False, null=False)
    received_value = models.DecimalField(
        max_digits=9, decimal_places=2, blank=True, null=True
    )
    liquid_value = models.DecimalField(
        max_digits=9, decimal_places=2, blank=True, null=True
    )
    payment_method = models.ForeignKey(
        PaymentMethod, on_delete=models.PROTECT, null=True
    )
    registry_classification = models.CharField(
        max_length=30,
        blank=True,
        null=False,
        default=RegistryClassification.VARIABLE,
        choices=RegistryClassification.CHOICES,
    )
    number_of_payments = models.PositiveIntegerField(default=1, null=False, blank=False)
    current_payment = models.PositiveIntegerField(default=1, null=False, blank=False)
    frequency = models.IntegerField(
        choices=Frequency.CHOICES, default=Frequency.NONE, null=True, blank=True
    )
    reference_record = models.ForeignKey(
        "self", on_delete=models.PROTECT, null=True, blank=False, related_name="payments"
    )
    observation = models.TextField(blank=True, null=True)
    cost_center = models.JSONField(default=dict)
    account = models.ForeignKey(Account, on_delete=models.PROTECT, blank=True, null=True)
    entry_type = models.CharField(max_length=2, choices=Type.CHOICES, blank=False)
    person = models.JSONField(default=dict)
    tag = models.TextField(blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    client_app = models.JSONField(default=dict, null=False, blank=False)



import operator
from functools import reduce
from rest_framework import viewsets
from rest_framework.views import APIView

from django.db.models.query_utils import Q

from models.bank import Bank
from serializers.bank import BankSerializer

class BankViewSet(viewsets.ModelViewSet):

    serializer_class = BankSerializer
    queryset = Bank.objects.all()

    def get_queryset(self):
        queryset = super().get_queryset()

        query = Q(deactivate=False)

        if self.action == "list":
            name = str(self.request.query_params.get("name", "")).split(" ")

            if name != "":
                query &= reduce(
                    operator.__and__,
                    (Q(name__unaccent__icontains=word) for word in name),
                )

        return queryset.filter(query).order_by("-name")



from rest_framework import serializers
from models.payment import PaymentMethod, PaymentMethodTaxs


class PaymentMethodSerializer(serializers.ModelSerializer):
    class Meta:
        model = PaymentMethod
        fields = ["id", "name", "tax"]

    def validate_name(self, value):
        if not (self.instance and value == self.instance.name):
            if PaymentMethod.objects.filter(
                name=value, deactivate=False, client_app=self.initial_data.get('client_app')).exists():
                raise serializers.ValidationError(
                    "Already exists a payment method with this name."
                )

        return value

    def validate_tax(self, value):
        if  value is not None and value < 0:
            raise serializers.ValidationError("the tax can't be negative")

        return value


class PaymentMethodTaxsSerializer(serializers.ModelSerializer):
    class Meta:
        model = PaymentMethodTaxs
        fields = ["id", "initial_payment", "final_payment", "tax"]

    def validate(self, attrs):
        if attrs["initial_payment"] > attrs["final_payment"]:
            raise serializers.ValidationError(
                {"menssage": " The initial payment can't be bigger then the final_payment"}
            )

        taxs = PaymentMethodTaxs.objects.filter(
            payment_method_id=self.context.get("payment_method_id"),
            client_app=self.initial_data.get('client_app'),
        )

        if self.instance:
            taxs = taxs.exclude(id=self.instance.id)

        for tax in taxs:
            if attrs["initial_payment"] in range(
                tax.initial_payment, tax.final_payment + 1
            ):
                raise serializers.ValidationError(
                    {
                        "initial_payment": "Already exists a tax for this payment in this payment method."
                    }
                )
            elif attrs["final_payment"] in range(
                tax.initial_payment, tax.final_payment + 1
            ):
                raise serializers.ValidationError(
                    {
                        "final_payment": "Already exists a tax for this payment in this payment method."
                    }
                )
            elif (
                tax.initial_payment >= attrs["initial_payment"]
                and tax.final_payment <= attrs["final_payment"]
            ):
                raise serializers.ValidationError(
                    {
                        "menssage": "Already exists a tax for this payments in this payment method."
                    }
                )

        return attrs


import operator
from functools import reduce
from rest_framework import viewsets
from rest_framework.views import APIView

from django.db.models.query_utils import Q

from models.category import Category
from serializers.category import CategorySerializer


class CategoryViewSet(viewsets.ModelViewSet):
    serializer_class = CategorySerializer
    queryset = Category.objects.all()

    def get_queryset(self):
        queryset = super().get_queryset()

        query = Q()

        if self.action == "list":
            query_params = self.request.query_params

            name = str(query_params.get("name", "")).split(" ")

            if name != "":
                query &= reduce(
                    operator.__and__,
                    (Q(name__unaccent__icontains=word) for word in name),
                )

            deactivates = query_params.get("deactivate", "").lower() == "true"

            query &= Q(deactivate=deactivates)

        return queryset.filter(query).order_by("name")


from rest_framework import serializers
from models.accounts import Account, AccountPermission


class AccountSerializer(serializers.ModelSerializer):
    class Meta:
        model = Account
        fields = ["id", "name", "bank"]

    def validate_name(self, value):
        if not (self.instance and value == self.instance.name):
            if Account.objects.filter(
                name=value, deactivate=False, client_app=self.initial_data.get('client_app')).exists():
                raise serializers.ValidationError(
                    "Already exists an account with this name."
                )

        return value
    


class AccountPermissionSerializer(serializers.ModelSerializer):
    
    class Meta:
        model = AccountPermission
        fields = ["id", "account", "employee"]
        validators = []

    def validate(self, attrs):
        if (
            AccountPermission.objects.filter(
                employee__id=attrs["employee"],
                account=attrs["account"],
                deactivate=False,
                account__deactivate=False,
                account__bank__deactivate=False,
            ).exists()
        ):
            raise serializers.ValidationError(
                {"employee": "This employee already has permission"}
            )

        return attrs
    

